from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.qdrant.qdrant import (
    QdrantAccessConfig,
    QdrantConnectionConfig,
    QdrantUploader,
    QdrantUploaderConfig,
    QdrantUploadStager,
    QdrantUploadStagerConfig,
)

CONNECTOR_TYPE = "qdrant-local"


class LocalQdrantAccessConfig(QdrantAccessConfig):
    pass


class LocalQdrantConnectionConfig(QdrantConnectionConfig):
    path: str = Field(default=None, description="Persistence path for QdrantLocal.")
    access_config: Secret[LocalQdrantAccessConfig] = Field(
        default_factory=LocalQdrantAccessConfig, validate_default=True
    )

    def get_client_kwargs(self) -> dict:
        return {"path": self.path}


class LocalQdrantUploadStagerConfig(QdrantUploadStagerConfig):
    pass


@dataclass
class LocalQdrantUploadStager(QdrantUploadStager):
    upload_stager_config: LocalQdrantUploadStagerConfig


class LocalQdrantUploaderConfig(QdrantUploaderConfig):
    pass


@dataclass
class LocalQdrantUploader(QdrantUploader):
    connection_config: LocalQdrantConnectionConfig
    upload_config: LocalQdrantUploaderConfig
    connector_type: str = CONNECTOR_TYPE


qdrant_local_destination_entry = DestinationRegistryEntry(
    connection_config=LocalQdrantConnectionConfig,
    uploader=LocalQdrantUploader,
    uploader_config=LocalQdrantUploaderConfig,
    upload_stager=LocalQdrantUploadStager,
    upload_stager_config=LocalQdrantUploadStagerConfig,
)



from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.utils.data_prep import (
    flatten_dict,
    get_data_df,
    get_enhanced_element_id,
    split_dataframe,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from kdbai_client import Database, Session, Table
    from pandas import DataFrame

CONNECTOR_TYPE = "kdbai"


class KdbaiAccessConfig(AccessConfig):
    api_key: Optional[str] = Field(
        default=None,
        description="A string for the api-key, can be left empty "
        "when connecting to local KDBAI instance.",
    )


class KdbaiConnectionConfig(ConnectionConfig):
    access_config: Secret[KdbaiAccessConfig] = Field(
        default=KdbaiAccessConfig(), validate_default=True
    )
    endpoint: str = Field(
        default="http://localhost:8082", description="Endpoint url where KDBAI is hosted."
    )

    @requires_dependencies(["kdbai_client"], extras="kdbai")
    @contextmanager
    def get_client(self) -> Generator["Session", None, None]:
        from kdbai_client import Session

        session = None
        try:
            session = Session(
                api_key=self.access_config.get_secret_value().api_key, endpoint=self.endpoint
            )
            yield session
        finally:
            if session:
                session.close()


class KdbaiUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class KdbaiUploadStager(UploadStager):
    upload_stager_config: KdbaiUploadStagerConfig = field(default_factory=KdbaiUploadStagerConfig)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        return {
            "id": get_enhanced_element_id(element_dict=data, file_data=file_data),
            "element_id": data.get("element_id"),
            "document": data.pop("text", None),
            "embeddings": data.get("embeddings"),
            "metadata": flatten_dict(
                dictionary=data.get("metadata"),
                flatten_lists=True,
                remove_none=True,
            ),
        }


class KdbaiUploaderConfig(UploaderConfig):
    database_name: str = Field(
        default="default", description="The name of the KDBAI database to write into."
    )
    table_name: str = Field(description="The name of the KDBAI table to write into.")
    batch_size: int = Field(default=100, description="Number of records per batch")


@dataclass
class KdbaiUploader(Uploader):
    connection_config: KdbaiConnectionConfig
    upload_config: KdbaiUploaderConfig
    connector_type: str = field(default=CONNECTOR_TYPE, init=False)

    def precheck(self) -> None:
        try:
            self.get_database()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @contextmanager
    def get_database(self) -> Generator["Database", None, None]:
        with self.connection_config.get_client() as client:
            db = client.database(self.upload_config.database_name)
            yield db

    @contextmanager
    def get_table(self) -> Generator["Table", None, None]:
        with self.get_database() as db:
            table = db.table(self.upload_config.table_name)
            yield table

    def upsert_batch(self, batch: "DataFrame"):
        with self.get_table() as table:
            table.insert(batch)

    def process_dataframe(self, df: "DataFrame"):
        logger.debug(
            f"uploading {len(df)} entries to {self.connection_config.endpoint} "
            f"db {self.upload_config.database_name} in table {self.upload_config.table_name}"
        )
        for batch_df in split_dataframe(df=df, chunk_size=self.upload_config.batch_size):
            self.upsert_batch(batch=batch_df)

    @requires_dependencies(["pandas"], extras="kdbai")
    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data=data)
        self.process_dataframe(df=df)

    @requires_dependencies(["pandas"], extras="kdbai")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        data = get_data_df(path=path)
        self.process_dataframe(df=data)


kdbai_destination_entry = DestinationRegistryEntry(
    connection_config=KdbaiConnectionConfig,
    uploader=KdbaiUploader,
    uploader_config=KdbaiUploaderConfig,
    upload_stager=KdbaiUploadStager,
    upload_stager_config=KdbaiUploadStagerConfig,
)



from typing import TYPE_CHECKING, Any

from unstructured_ingest.utils.data_prep import flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from pandas import DataFrame


@requires_dependencies(["pandas"])
def get_default_pandas_dtypes() -> dict[str, Any]:
    import pandas as pd

    return {
        "text": pd.StringDtype(),  # type: ignore
        "type": pd.StringDtype(),  # type: ignore
        "element_id": pd.StringDtype(),  # type: ignore
        "filename": pd.StringDtype(),  # Optional[str]  # type: ignore
        "filetype": pd.StringDtype(),  # Optional[str]  # type: ignore
        "file_directory": pd.StringDtype(),  # Optional[str]  # type: ignore
        "last_modified": pd.StringDtype(),  # Optional[str]  # type: ignore
        "attached_to_filename": pd.StringDtype(),  # Optional[str]  # type: ignore
        "parent_id": pd.StringDtype(),  # Optional[str],  # type: ignore
        "category_depth": "Int64",  # Optional[int]
        "image_path": pd.StringDtype(),  # Optional[str]  # type: ignore
        "languages": object,  # Optional[list[str]]
        "page_number": "Int64",  # Optional[int]
        "page_name": pd.StringDtype(),  # Optional[str]  # type: ignore
        "url": pd.StringDtype(),  # Optional[str]  # type: ignore
        "link_urls": pd.StringDtype(),  # Optional[str]  # type: ignore
        "link_texts": object,  # Optional[list[str]]
        "links": object,
        "sent_from": object,  # Optional[list[str]],
        "sent_to": object,  # Optional[list[str]]
        "subject": pd.StringDtype(),  # Optional[str]  # type: ignore
        "section": pd.StringDtype(),  # Optional[str]  # type: ignore
        "header_footer_type": pd.StringDtype(),  # Optional[str]  # type: ignore
        "emphasized_text_contents": object,  # Optional[list[str]]
        "emphasized_text_tags": object,  # Optional[list[str]]
        "text_as_html": pd.StringDtype(),  # Optional[str]  # type: ignore
        "regex_metadata": object,
        "max_characters": "Int64",  # Optional[int]
        "is_continuation": "boolean",  # Optional[bool]
        "detection_class_prob": float,  # Optional[float],
        "sender": pd.StringDtype(),  # type: ignore
        "coordinates_points": object,
        "coordinates_system": pd.StringDtype(),  # type: ignore
        "coordinates_layout_width": float,
        "coordinates_layout_height": float,
        "data_source_url": pd.StringDtype(),  # Optional[str]  # type: ignore
        "data_source_version": pd.StringDtype(),  # Optional[str]  # type: ignore
        "data_source_record_locator": object,
        "data_source_date_created": pd.StringDtype(),  # Optional[str]  # type: ignore
        "data_source_date_modified": pd.StringDtype(),  # Optional[str]  # type: ignore
        "data_source_date_processed": pd.StringDtype(),  # Optional[str]  # type: ignore
        "data_source_permissions_data": object,
        "embeddings": object,
        "regex_metadata_key": object,
    }


def convert_to_pandas_dataframe(
    elements_dict: list[dict[str, Any]],
    drop_empty_cols: bool = False,
) -> "DataFrame":
    import pandas as pd

    # Flatten metadata if it hasn't already been flattened
    for d in elements_dict:
        if metadata := d.pop("metadata", None):
            d.update(flatten_dict(metadata, keys_to_omit=["data_source_record_locator"]))

    df = pd.DataFrame.from_dict(
        elements_dict,
    )
    dt = {k: v for k, v in get_default_pandas_dtypes().items() if k in df.columns}
    df = df.astype(dt)
    if drop_empty_cols:
        df.dropna(axis=1, how="all", inplace=True)
    return df



# https://developers.notion.com/reference/property-object#phone-number
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase


@dataclass
class PhoneNumber(DBPropertyBase):
    id: str
    name: str
    type: str = "phone_number"
    description: Optional[str] = None
    phone_number: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class PhoneNumberCell(DBCellBase):
    id: str
    phone_number: Optional[str]
    name: Optional[str] = None
    type: str = "phone_number"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        if phone_number := self.phone_number:
            return Div([], phone_number)
        return None



import datetime as dt
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import UserAuthError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import SourceRegistryEntry
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from discord import Client as DiscordClient

CONNECTOR_TYPE = "discord"


class DiscordAccessConfig(AccessConfig):
    token: str = Field(description="Discord API token")


class DiscordConnectionConfig(ConnectionConfig):
    access_config: Secret[DiscordAccessConfig] = Field(
        default=DiscordAccessConfig, validate_default=True
    )

    @requires_dependencies(["discord"], extras="discord")
    def get_client(self) -> "DiscordClient":
        import discord

        intents = discord.Intents.default()
        intents.message_content = True
        return discord.Client(intents=intents)


class DiscordIndexerConfig(IndexerConfig):
    channels: list[str] = Field(
        default=None,
        description="List of Discord channel IDs to process",
    )


@dataclass
class DiscordIndexer(Indexer):
    connection_config: DiscordConnectionConfig
    index_config: DiscordIndexerConfig

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        self.connection_config.get_client()
        channels_to_process: set[str] = set(self.index_config.channels or [])

        for channel_id in list(channels_to_process):
            file_data = self.get_channel_file_data(channel_id=channel_id)
            if file_data:
                yield file_data

    def precheck(self) -> None:
        if not self.connection_config.access_config.get_secret_value().token:
            raise UserAuthError("Discord token is missing")
        if not self.index_config.channels:
            raise ValueError("No channels provided")

    def get_channel_file_data(self, channel_id: str) -> Optional[FileData]:
        # Fetch channel metadata
        identifier = channel_id
        channel_id = f"{channel_id}.txt"
        source_identifiers = SourceIdentifiers(
            filename=channel_id,
            fullpath=channel_id,
        )
        metadata = FileDataSourceMetadata(
            record_locator={"channel_id": identifier},
            date_processed=str(dt.datetime.utcnow().isoformat()),
        )
        return FileData(
            identifier=identifier,
            connector_type=CONNECTOR_TYPE,
            source_identifiers=source_identifiers,
            metadata=metadata,
            display_name=source_identifiers.fullpath,
        )


class DiscordDownloaderConfig(DownloaderConfig):
    limit: Optional[int] = Field(
        default=100, description="Limit on how many messages per channel to pull in"
    )


@dataclass
class DiscordDownloader(Downloader):
    connection_config: DiscordConnectionConfig
    download_config: DiscordDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    def is_async(self) -> bool:
        return True

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        # Synchronous run is not implemented
        raise NotImplementedError()

    async def run_async(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        record_locator = file_data.metadata.record_locator

        if "channel_id" not in record_locator:
            raise ValueError(f"No channel id in file data record locator: {record_locator}")

        client = self.connection_config.get_client()
        download_path = self.get_download_path(file_data=file_data)
        download_path.parent.mkdir(parents=True, exist_ok=True)

        messages = []
        channel_id = record_locator["channel_id"]

        @client.event
        async def on_ready():
            logger.debug("Discord Bot is ready")
            channel = client.get_channel(int(channel_id))
            if not channel:
                raise ValueError(f"channel not found for id: {channel_id}")
            logger.debug(f"Processing messages for channel: {channel.name}")
            async for msg in channel.history(limit=self.download_config.limit):
                messages.append(msg)
            logger.debug(f"Fetched {len(messages)} messages")
            await client.close()

        try:
            await client.start(self.connection_config.access_config.get_secret_value().token)
        finally:
            await client.close()

        content = "\n".join([message.content for message in messages])

        with open(download_path, "w") as file:
            file.write(content)

        return self.generate_download_response(file_data=file_data, download_path=download_path)


discord_source_entry = SourceRegistryEntry(
    indexer=DiscordIndexer,
    indexer_config=DiscordIndexerConfig,
    downloader=DiscordDownloader,
    downloader_config=DiscordDownloaderConfig,
    connection_config=DiscordConnectionConfig,
)



from dataclasses import dataclass, field
from datetime import date, datetime
from typing import TYPE_CHECKING, Annotated, Any, Optional

from dateutil import parser
from pydantic import Field, Secret
from pydantic.functional_validators import BeforeValidator

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.utils.data_prep import (
    batch_generator,
    flatten_dict,
    get_enhanced_element_id,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

from .utils import conform_string_to_dict

if TYPE_CHECKING:
    from chromadb import Client

CONNECTOR_TYPE = "chroma"


class ChromaAccessConfig(AccessConfig):
    settings: Optional[Annotated[dict, BeforeValidator(conform_string_to_dict)]] = Field(
        default=None, description="A dictionary of settings to communicate with the chroma server."
    )
    headers: Optional[Annotated[dict, BeforeValidator(conform_string_to_dict)]] = Field(
        default=None, description="A dictionary of headers to send to the Chroma server."
    )


class ChromaConnectionConfig(ConnectionConfig):
    access_config: Secret[ChromaAccessConfig] = Field(
        default=ChromaAccessConfig(), validate_default=True
    )
    path: Optional[str] = Field(
        default=None, description="Location where Chroma is persisted, if not connecting via http."
    )
    tenant: Optional[str] = Field(
        default="default_tenant", description="The tenant to use for this client."
    )
    database: Optional[str] = Field(
        default="default_database", description="The database to use for this client."
    )
    host: Optional[str] = Field(default=None, description="The hostname of the Chroma server.")
    port: Optional[int] = Field(default=None, description="The port of the Chroma server.")
    ssl: bool = Field(
        default=False, description="Whether to use SSL to connect to the Chroma server."
    )
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @requires_dependencies(["chromadb"], extras="chroma")
    def get_client(self) -> "Client":
        import chromadb

        access_config = self.access_config.get_secret_value()
        if path := self.path:
            return chromadb.PersistentClient(
                path=path,
                settings=access_config.settings,
                tenant=self.tenant,
                database=self.database,
            )

        elif (host := self.host) and (port := self.port):
            return chromadb.HttpClient(
                host=host,
                port=str(port),
                ssl=self.ssl,
                headers=access_config.headers,
                settings=access_config.settings,
                tenant=self.tenant,
                database=self.database,
            )
        else:
            raise ValueError("Chroma connector requires either path or host and port to be set.")


class ChromaUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class ChromaUploadStager(UploadStager):
    upload_stager_config: ChromaUploadStagerConfig = field(
        default_factory=lambda: ChromaUploadStagerConfig()
    )

    @staticmethod
    def parse_date_string(date_string: str) -> date:
        try:
            timestamp = float(date_string)
            return datetime.fromtimestamp(timestamp)
        except Exception as e:
            logger.debug(f"date {date_string} string not a timestamp: {e}")
        return parser.parse(date_string)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        """
        Prepares dictionary in the format that Chroma requires
        """
        data = element_dict.copy()
        return {
            "id": get_enhanced_element_id(element_dict=data, file_data=file_data),
            "embedding": data.pop("embeddings", None),
            "document": data.pop("text", None),
            "metadata": flatten_dict(data, separator="-", flatten_lists=True, remove_none=True),
        }


class ChromaUploaderConfig(UploaderConfig):
    collection_name: str = Field(description="The name of the Chroma collection to write into.")
    batch_size: int = Field(default=100, description="Number of records per batch")


@dataclass
class ChromaUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: ChromaUploaderConfig
    connection_config: ChromaConnectionConfig

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @DestinationConnectionError.wrap
    def upsert_batch(self, collection, batch):
        try:
            # Chroma wants lists even if there is only one element
            # Upserting to prevent duplicates
            collection.upsert(
                ids=batch["ids"],
                documents=batch["documents"],
                embeddings=batch["embeddings"],
                metadatas=batch["metadatas"],
            )
        except Exception as e:
            raise DestinationConnectionError(f"chroma error: {e}") from e

    @staticmethod
    def prepare_chroma_list(chunk: tuple[dict[str, Any]]) -> dict[str, list[Any]]:
        """Helper function to break a tuple of dicts into list of parallel lists for ChromaDb.
        ({'id':1}, {'id':2}, {'id':3}) -> {'ids':[1,2,3]}"""
        chroma_dict = {}
        chroma_dict["ids"] = [x.get("id") for x in chunk]
        chroma_dict["documents"] = [x.get("document") for x in chunk]
        chroma_dict["embeddings"] = [x.get("embedding") for x in chunk]
        chroma_dict["metadatas"] = [x.get("metadata") for x in chunk]
        # Make sure all lists are of the same length
        assert (
            len(chroma_dict["ids"])
            == len(chroma_dict["documents"])
            == len(chroma_dict["embeddings"])
            == len(chroma_dict["metadatas"])
        )
        return chroma_dict

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing {len(data)} objects to destination "
            f"collection {self.upload_config.collection_name} "
            f"at {self.connection_config.host}",
        )
        client = self.connection_config.get_client()

        collection = client.get_or_create_collection(name=self.upload_config.collection_name)
        for chunk in batch_generator(data, self.upload_config.batch_size):
            self.upsert_batch(collection, self.prepare_chroma_list(chunk))


chroma_destination_entry = DestinationRegistryEntry(
    connection_config=ChromaConnectionConfig,
    uploader=ChromaUploader,
    uploader_config=ChromaUploaderConfig,
    upload_stager=ChromaUploadStager,
    upload_stager_config=ChromaUploadStagerConfig,
)



import asyncio
import json
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Mapping, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.utils.data_prep import flatten_dict
from unstructured_ingest.utils.dep_check import requires_dependencies

BASE_URL = "https://api.vectara.io/v2"

CONNECTOR_TYPE = "vectara"


class VectaraAccessConfig(AccessConfig):
    oauth_client_id: str = Field(description="Client ID")
    oauth_secret: str = Field(description="Client Secret")


class VectaraConnectionConfig(ConnectionConfig):
    access_config: Secret[VectaraAccessConfig]
    customer_id: str
    corpus_name: Optional[str] = None
    corpus_key: Optional[str] = None
    token_url: str = "https://vectara-prod-{}.auth.us-west-2.amazoncognito.com/oauth2/token"


class VectaraUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class VectaraUploadStager(UploadStager):
    upload_stager_config: VectaraUploadStagerConfig = field(
        default_factory=lambda: VectaraUploadStagerConfig()
    )

    @staticmethod
    def conform_dict(data: dict) -> dict:
        """
        Prepares dictionary in the format that Vectara requires.
        See more detail in https://docs.vectara.com/docs/rest-api/create-corpus-document

        Select which meta-data fields to include and optionally map them to a new format.
        remove the "metadata-" prefix from the keys
        """
        metadata_map = {
            "page_number": "page_number",
            "data_source-url": "url",
            "filename": "filename",
            "filetype": "filetype",
            "last_modified": "last_modified",
            "element_id": "element_id",
        }
        md = flatten_dict(data, separator="-", flatten_lists=True)
        md = {k.replace("metadata-", ""): v for k, v in md.items()}
        md = {metadata_map[k]: v for k, v in md.items() if k in metadata_map}
        return md

    def process_whole(self, input_file: Path, output_file: Path, file_data: FileData) -> None:
        with input_file.open() as in_f:
            elements_contents = json.load(in_f)

        logger.info(
            f"Extending {len(elements_contents)} json elements from content in {input_file}"
        )

        conformed_elements = [
            {
                "id": str(uuid.uuid4()),
                "type": "core",
                "metadata": {
                    "title": file_data.identifier,
                },
                "document_parts": [
                    {
                        "text": element.pop("text", None),
                        "metadata": self.conform_dict(data=element),
                    }
                    for element in elements_contents
                ],
            }
        ]

        with open(output_file, "w") as out_f:
            json.dump(conformed_elements, out_f, indent=2)


class VectaraUploaderConfig(UploaderConfig):
    pass


@dataclass
class VectaraUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: VectaraUploaderConfig
    connection_config: VectaraConnectionConfig
    _jwt_token: Optional[str] = field(init=False, default=None)
    _jwt_token_expires_ts: Optional[float] = field(init=False, default=None)

    def is_async(self) -> bool:
        return True

    def precheck(self) -> None:
        try:
            self._check_connection_and_corpora()
        except Exception as e:
            logger.error(f"Failed to validate connection {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @property
    async def jwt_token_async(self) -> str:
        if not self._jwt_token or self._jwt_token_expires_ts - datetime.now().timestamp() <= 60:
            self._jwt_token = await self._get_jwt_token_async()
        return self._jwt_token

    @property
    def jwt_token(self) -> str:
        if not self._jwt_token or self._jwt_token_expires_ts - datetime.now().timestamp() <= 60:
            self._jwt_token = self._get_jwt_token()
        return self._jwt_token

    # Get Oauth2 JWT token
    @requires_dependencies(["httpx"], extras="vectara")
    async def _get_jwt_token_async(self) -> str:
        import httpx

        """Connect to the server and get a JWT token."""
        token_endpoint = self.connection_config.token_url.format(self.connection_config.customer_id)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "grant_type": "client_credentials",
            "client_id": self.connection_config.access_config.get_secret_value().oauth_client_id,
            "client_secret": self.connection_config.access_config.get_secret_value().oauth_secret,
        }

        async with httpx.AsyncClient() as client:
            response = await client.post(token_endpoint, headers=headers, data=data)
            response.raise_for_status()
            response_json = response.json()

        request_time = datetime.now().timestamp()
        self._jwt_token_expires_ts = request_time + response_json.get("expires_in")

        return response_json.get("access_token")

    # Get Oauth2 JWT token
    @requires_dependencies(["httpx"], extras="vectara")
    def _get_jwt_token(self) -> str:
        import httpx

        """Connect to the server and get a JWT token."""
        token_endpoint = self.connection_config.token_url.format(self.connection_config.customer_id)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {
            "grant_type": "client_credentials",
            "client_id": self.connection_config.access_config.get_secret_value().oauth_client_id,
            "client_secret": self.connection_config.access_config.get_secret_value().oauth_secret,
        }

        with httpx.Client() as client:
            response = client.post(token_endpoint, headers=headers, data=data)
            response.raise_for_status()
            response_json = response.json()

        request_time = datetime.now().timestamp()
        self._jwt_token_expires_ts = request_time + response_json.get("expires_in")

        return response_json.get("access_token")

    @DestinationConnectionError.wrap
    def _check_connection_and_corpora(self) -> None:
        """
        Check the connection for Vectara and validate corpus exists.
        - If more than one corpus with the same name exists - raise error
        - If exactly one corpus exists with this name - use it.
        - If does not exist - raise error.
        """
        # Get token if not already set
        self.jwt_token

        _, list_corpora_response = self._request(
            http_method="GET",
            endpoint="corpora",
        )

        if self.connection_config.corpus_name:
            possible_corpora_keys_names_map = {
                corpus.get("key"): corpus.get("name")
                for corpus in list_corpora_response.get("corpora")
                if corpus.get("name") == self.connection_config.corpus_name
            }

            if len(possible_corpora_keys_names_map) > 1:
                raise ValueError(
                    f"Multiple Corpus exist with name {self.connection_config.corpus_name} in dest."
                )
            if len(possible_corpora_keys_names_map) == 1:
                if not self.connection_config.corpus_key:
                    self.connection_config.corpus_key = list(
                        possible_corpora_keys_names_map.keys()
                    )[0]
                elif (
                    self.connection_config.corpus_key
                    != list(possible_corpora_keys_names_map.keys())[0]
                ):
                    raise ValueError("Corpus key does not match provided corpus name.")
            else:
                raise ValueError(
                    f"No Corpora exist with name {self.connection_config.corpus_name} in dest."
                )

    @requires_dependencies(["httpx"], extras="vectara")
    async def _async_request(
        self,
        endpoint: str,
        http_method: str = "POST",
        params: Mapping[str, Any] = None,
        data: Mapping[str, Any] = None,
    ) -> tuple[bool, dict]:
        import httpx

        url = f"{BASE_URL}/{endpoint}"

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {await self.jwt_token_async}",
            "X-source": "unstructured",
        }

        async with httpx.AsyncClient() as client:
            response = await client.request(
                method=http_method, url=url, headers=headers, params=params, json=data
            )
            response.raise_for_status()
            return response.json()

    @requires_dependencies(["httpx"], extras="vectara")
    def _request(
        self,
        endpoint: str,
        http_method: str = "POST",
        params: Mapping[str, Any] = None,
        data: Mapping[str, Any] = None,
    ) -> tuple[bool, dict]:
        import httpx

        url = f"{BASE_URL}/{endpoint}"

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.jwt_token}",
            "X-source": "unstructured",
        }

        with httpx.Client() as client:
            response = client.request(
                method=http_method, url=url, headers=headers, params=params, json=data
            )
            response.raise_for_status()
            return response.json()

    async def _delete_doc(self, doc_id: str) -> tuple[bool, dict]:
        """
        Delete a document from the Vectara corpus.
        """

        return await self._async_request(
            endpoint=f"corpora/{self.connection_config.corpus_key}/documents/{doc_id}",
            http_method="DELETE",
        )

    async def _index_document(self, document: Dict[str, Any]) -> None:
        """
        Index a document (by uploading it to the Vectara corpus) from the document dictionary
        """

        logger.debug(
            f"Indexing document {document['id']} to corpus key {self.connection_config.corpus_key}"
        )

        try:
            result = await self._async_request(
                endpoint=f"corpora/{self.connection_config.corpus_key}/documents", data=document
            )
        except Exception as e:
            logger.error(f"exception {e} while indexing document {document['id']}")
            return

        if (
            "messages" in result
            and result["messages"]
            and (
                "ALREADY_EXISTS" in result["messages"]
                or (
                    "CONFLICT: Indexing doesn't support updating documents."
                    in result["messages"][0]
                )
            )
        ):
            logger.info(f"document {document['id']} already exists, re-indexing")
            await self._delete_doc(document["id"])
            await self._async_request(
                endpoint=f"corpora/{self.connection_config.corpus_key}/documents", data=document
            )
            return

        logger.info(f"indexing document {document['id']} succeeded")

    async def run_data_async(
        self,
        data: list[dict],
        file_data: FileData,
        **kwargs: Any,
    ) -> None:
        logger.info(f"inserting / updating {len(data)} documents to Vectara ")
        await asyncio.gather(*(self._index_document(vdoc) for vdoc in data))


vectara_destination_entry = DestinationRegistryEntry(
    connection_config=VectaraConnectionConfig,
    uploader=VectaraUploader,
    uploader_config=VectaraUploaderConfig,
    upload_stager=VectaraUploadStager,
    upload_stager_config=VectaraUploadStagerConfig,
)



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.error import (
    ProviderError,
    QuotaError,
    RateLimitError,
    UserAuthError,
    UserError,
    is_internal_error,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from openai import AsyncOpenAI, OpenAI


class OctoAiEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr = Field(description="API key for OctoAI")
    embedder_model_name: str = Field(
        default="thenlper/gte-large", alias="model_name", description="octoai model name"
    )
    base_url: str = Field(
        default="https://text.octoai.run/v1", description="optional override for the base url"
    )

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        # https://platform.openai.com/docs/guides/error-codes/api-errors
        from openai import APIStatusError

        if not isinstance(e, APIStatusError):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            raise e
        error_code = e.code
        if 400 <= e.status_code < 500:
            # user error
            if e.status_code == 401:
                return UserAuthError(e.message)
            if e.status_code == 429:
                # 429 indicates rate limit exceeded and quote exceeded
                if error_code == "insufficient_quota":
                    return QuotaError(e.message)
                else:
                    return RateLimitError(e.message)
            return UserError(e.message)
        if e.status_code >= 500:
            return ProviderError(e.message)
        logger.error(f"unhandled exception from openai: {e}", exc_info=True)
        return e

    def run_precheck(self) -> None:
        client = self.get_client()
        try:
            models = [m.id for m in list(client.models.list())]
            if self.embedder_model_name not in models:
                raise UserError(
                    "model '{}' not found: {}".format(self.embedder_model_name, ", ".join(models))
                )
        except Exception as e:
            raise self.wrap_error(e=e)

    @requires_dependencies(
        ["openai", "tiktoken"],
        extras="octoai",
    )
    def get_client(self) -> "OpenAI":
        """Creates an OpenAI python client to embed elements. Uses the OpenAI SDK."""
        from openai import OpenAI

        return OpenAI(api_key=self.api_key.get_secret_value(), base_url=self.base_url)

    @requires_dependencies(
        ["openai", "tiktoken"],
        extras="octoai",
    )
    def get_async_client(self) -> "AsyncOpenAI":
        """Creates an OpenAI python client to embed elements. Uses the OpenAI SDK."""
        from openai import AsyncOpenAI

        return AsyncOpenAI(api_key=self.api_key.get_secret_value(), base_url=self.base_url)


@dataclass
class OctoAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: OctoAiEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def _embed_query(self, query: str):
        client = self.get_client()
        response = client.embeddings.create(input=query, model=self.config.embedder_model_name)
        return response.data[0].embedding

    def get_client(self) -> "OpenAI":
        return self.config.get_client()

    def embed_batch(self, client: "OpenAI", batch: list[str]) -> list[list[float]]:
        response = client.embeddings.create(input=batch, model=self.config.embedder_model_name)
        return [data.embedding for data in response.data]


@dataclass
class AsyncOctoAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: OctoAiEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "AsyncOpenAI":
        return self.config.get_async_client()

    async def embed_batch(self, client: "AsyncOpenAI", batch: list[str]) -> list[list[float]]:
        response = await client.embeddings.create(
            input=batch, model=self.config.embedder_model_name
        )
        return [data.embedding for data in response.data]



"""
Salesforce Connector
Able to download Account, Case, Campaign, EmailMessage, Lead
Salesforce returns everything as a list of json.
This saves each entry as a separate file to be partitioned.
Using JWT authorization
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_key_and_cert.htm
https://developer.salesforce.com/docs/atlas.en-us.sfdx_dev.meta/sfdx_dev/sfdx_dev_auth_connected_app.htm
"""

import json
from collections import OrderedDict
from dataclasses import dataclass, field
from email.utils import formatdate
from pathlib import Path
from string import Template
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Generator, Optional, Type

from dateutil import parser
from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import (
    MissingCategoryError,
    SourceConnectionError,
    SourceConnectionNetworkError,
    ValueError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

CONNECTOR_TYPE = "salesforce"

if TYPE_CHECKING:
    from simple_salesforce import Salesforce

SALESFORCE_API_VERSION = "57.0"

# TODO: Add more categories as needed
ACCEPTED_CATEGORIES: list[str] = ["Account", "Case", "Campaign", "EmailMessage", "Lead"]

# Generic minimal email template used only
# to process EmailMessage records as .eml files
EMAIL_TEMPLATE = Template(
    """MIME-Version: 1.0
Date: $date
Message-ID: $message_identifier
Subject: $subject
From: $from_email
To: $to_email
Content-Type: multipart/alternative; boundary="00000000000095c9b205eff92630"
--00000000000095c9b205eff92630
Content-Type: text/plain; charset="UTF-8"
$textbody
--00000000000095c9b205eff92630
Content-Type: text/html; charset="UTF-8"
$htmlbody
--00000000000095c9b205eff92630--
""",
)


class SalesforceAccessConfig(AccessConfig):
    consumer_key: str
    private_key_path: Optional[Path] = Field(
        default=None,
        description="Path to the private key file. Key file is usually named server.key.",
    )
    private_key: Optional[str] = Field(default=None, description="Contents of the private key")

    def model_post_init(self, __context: Any) -> None:
        if self.private_key_path is None and self.private_key is None:
            raise ValueError("either private_key or private_key_path must be set")
        if self.private_key is not None and self.private_key_path is not None:
            raise ValueError("only one of private_key or private_key_path must be set")

    @requires_dependencies(["cryptography"])
    def get_private_key_value_and_type(self) -> tuple[str, Type]:
        from cryptography.hazmat.primitives import serialization

        if self.private_key_path and self.private_key_path.is_file():
            return str(self.private_key_path), Path
        if self.private_key:
            try:
                serialization.load_pem_private_key(
                    data=str(self.private_key).encode("utf-8"), password=None
                )
            except Exception as e:
                raise ValueError(f"failed to validate private key data: {e}") from e
            return self.private_key, str

        raise ValueError("private_key does not contain PEM private key or path")


class SalesforceConnectionConfig(ConnectionConfig):
    username: str
    access_config: Secret[SalesforceAccessConfig]

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def get_client(self) -> "Salesforce":
        from simple_salesforce import Salesforce

        access_config = self.access_config.get_secret_value()
        pkey_value, pkey_type = access_config.get_private_key_value_and_type()

        return Salesforce(
            username=self.username,
            consumer_key=access_config.consumer_key,
            privatekey_file=pkey_value if pkey_type is Path else None,
            privatekey=pkey_value if pkey_type is str else None,
            version=SALESFORCE_API_VERSION,
        )


class SalesforceIndexerConfig(IndexerConfig):
    categories: list[str]


@dataclass
class SalesforceIndexer(Indexer):
    connection_config: SalesforceConnectionConfig
    index_config: SalesforceIndexerConfig

    def __post_init__(self):
        for record_type in self.index_config.categories:
            if record_type not in ACCEPTED_CATEGORIES:
                raise ValueError(f"{record_type} not currently an accepted Salesforce category")

    def precheck(self) -> None:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def get_file_extension(self, record_type) -> str:
        if record_type == "EmailMessage":
            extension = ".eml"
        elif record_type in ["Account", "Lead", "Case", "Campaign"]:
            extension = ".xml"
        else:
            raise MissingCategoryError(
                f"There are no categories with the name: {record_type}",
            )
        return extension

    @requires_dependencies(["simple_salesforce"], extras="salesforce")
    def list_files(self) -> list[FileData]:
        """Get Salesforce Ids for the records.
        Send them to next phase where each doc gets downloaded into the
        appropriate format for partitioning.
        """
        from simple_salesforce.exceptions import SalesforceMalformedRequest

        client = self.connection_config.get_client()

        files_list = []
        for record_type in self.index_config.categories:
            try:
                # Get ids from Salesforce
                records = client.query_all_iter(
                    f"select Id, SystemModstamp, CreatedDate, LastModifiedDate from {record_type}",
                )
                for record in records:
                    record_with_extension = record["Id"] + self.get_file_extension(
                        record["attributes"]["type"]
                    )
                    source_identifiers = SourceIdentifiers(
                        filename=record_with_extension,
                        fullpath=f"{record['attributes']['type']}/{record_with_extension}",
                    )
                    files_list.append(
                        FileData(
                            connector_type=CONNECTOR_TYPE,
                            identifier=record["Id"],
                            source_identifiers=source_identifiers,
                            metadata=FileDataSourceMetadata(
                                url=record["attributes"]["url"],
                                version=str(parser.parse(record["SystemModstamp"]).timestamp()),
                                date_created=str(parser.parse(record["CreatedDate"]).timestamp()),
                                date_modified=str(
                                    parser.parse(record["LastModifiedDate"]).timestamp()
                                ),
                                record_locator={"id": record["Id"]},
                            ),
                            additional_metadata={"record_type": record["attributes"]["type"]},
                            display_name=source_identifiers.fullpath,
                        )
                    )
            except SalesforceMalformedRequest as e:
                raise SalesforceMalformedRequest(f"Problem with Salesforce query: {e}")

        return files_list

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for f in self.list_files():
            yield f


class SalesforceDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class SalesforceDownloader(Downloader):
    connection_config: SalesforceConnectionConfig
    download_config: SalesforceDownloaderConfig = field(
        default_factory=lambda: SalesforceDownloaderConfig()
    )
    connector_type: str = CONNECTOR_TYPE

    def _xml_for_record(self, record: OrderedDict) -> str:
        """Creates partitionable xml file from a record"""
        import xml.etree.ElementTree as ET

        def create_xml_doc(data, parent, prefix=""):
            for key, value in data.items():
                if isinstance(value, OrderedDict):
                    create_xml_doc(value, parent, prefix=f"{prefix}{key}.")
                else:
                    item = ET.Element("item")
                    item.text = f"{prefix}{key}: {value}"
                    parent.append(item)

        root = ET.Element("root")
        create_xml_doc(record, root)

        xml_string = ET.tostring(root, encoding="utf-8", xml_declaration=True).decode()
        return xml_string

    def _eml_for_record(self, email_json: dict[str, Any]) -> str:
        """Recreates standard expected .eml format using template."""
        eml = EMAIL_TEMPLATE.substitute(
            date=formatdate(parser.parse(email_json.get("MessageDate")).timestamp()),
            message_identifier=email_json.get("MessageIdentifier"),
            subject=email_json.get("Subject"),
            from_email=email_json.get("FromAddress"),
            to_email=email_json.get("ToAddress"),
            textbody=email_json.get("TextBody"),
            htmlbody=email_json.get("HtmlBody"),
        )
        return dedent(eml)

    @SourceConnectionNetworkError.wrap
    def _get_response(self, file_data: FileData) -> OrderedDict:
        client = self.connection_config.get_client()
        return client.query(
            f"select FIELDS(STANDARD) from {file_data.additional_metadata['record_type']} where Id='{file_data.identifier}'",  # noqa: E501
        )

    def get_record(self, file_data: FileData) -> OrderedDict:
        # Get record from Salesforce based on id
        response = self._get_response(file_data)
        logger.debug(f"response was returned for salesforce record id: {file_data.identifier}")
        records = response["records"]
        if not records:
            raise ValueError(
                f"No record found with record id {file_data.identifier}: {json.dumps(response)}"
            )
        record_json = records[0]
        return record_json

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        record = self.get_record(file_data)

        try:
            if file_data.additional_metadata["record_type"] == "EmailMessage":
                document = self._eml_for_record(record)
            else:
                document = self._xml_for_record(record)
            download_path = self.get_download_path(file_data=file_data)
            download_path.parent.mkdir(parents=True, exist_ok=True)

            with open(download_path, "w") as page_file:
                page_file.write(document)

        except Exception as e:
            logger.error(f"failed to download file {file_data.identifier}: {e}", exc_info=True)
            raise SourceConnectionNetworkError(f"failed to download file {file_data.identifier}")

        return self.generate_download_response(file_data=file_data, download_path=download_path)


salesforce_source_entry = SourceRegistryEntry(
    connection_config=SalesforceConnectionConfig,
    indexer_config=SalesforceIndexerConfig,
    indexer=SalesforceIndexer,
    downloader_config=SalesforceDownloaderConfig,
    downloader=SalesforceDownloader,
)



import logging
import os
from dataclasses import dataclass, field
from typing import Callable, ClassVar, Optional, Protocol, Sequence

from opentelemetry import trace
from opentelemetry.context import attach, get_current
from opentelemetry.propagate import extract, inject
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import ReadableSpan, Tracer, TracerProvider
from opentelemetry.sdk.trace.export import (
    ConsoleSpanExporter,
    SimpleSpanProcessor,
    SpanExportResult,
)

from unstructured_ingest.logger import logger


class AddTraceCallable(Protocol):
    def __call__(self, provider: TracerProvider) -> None:
        pass


class LogSpanExporter(ConsoleSpanExporter):
    def __init__(self, log_out: Callable = logger.info, **kwargs):
        self.log_out = log_out
        super().__init__(**kwargs)

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for span in spans:
            self.log_out(self.formatter(span))
        return SpanExportResult.SUCCESS


def get_log_out() -> Callable:
    level_names_mapping = {
        "CRITICAL": logging.CRITICAL,
        "FATAL": logging.FATAL,
        "ERROR": logging.ERROR,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }
    log_level = os.getenv("OTEL_LOG_LEVEL", "DEBUG").upper()
    log_level_int = level_names_mapping.get(log_level, logging.DEBUG)
    return lambda message: logger.log(log_level_int, message)


@dataclass
class OtelHandler:
    otel_endpoint: Optional[str] = None
    service_name: str = "unstructured-ingest"
    trace_provider: TracerProvider = field(init=False)
    log_out: Callable = field(default=get_log_out())
    trace_context_key: ClassVar[str] = "_trace_context"

    def init_trace(self):
        # Should only be done once
        resource = Resource(attributes={SERVICE_NAME: self.service_name})
        trace_provider = self.init_trace_provider(resource=resource)
        trace.set_tracer_provider(trace_provider)

    @staticmethod
    def set_attributes(span, attributes_dict):
        if attributes_dict:
            for att in attributes_dict:
                span.set_attribute(att, attributes_dict[att])

    @staticmethod
    def inject_context() -> dict:
        trace_context = {}
        current_context = get_current()
        inject(trace_context, current_context)
        return trace_context

    @staticmethod
    def attach_context(trace_context: dict) -> object:
        extracted_context = extract(trace_context)
        return attach(extracted_context)

    def get_otel_endpoint(self) -> Optional[str]:
        if otel_endpoint := self.otel_endpoint:
            return otel_endpoint
        if otlp_endpoint := os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT"):
            return otlp_endpoint
        if otlp_traces_endpoint := os.getenv("OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"):
            return otlp_traces_endpoint
        return None

    def _add_console_trace_processor(self, provider: TracerProvider) -> None:
        def custom_formatter(span: ReadableSpan) -> str:
            duration = (span.end_time - span.start_time) / 1e9
            s = f"{span.name} finished in {duration}s"
            if span.attributes:
                attributes_str = ", ".join([f"{k}={v}" for k, v in span.attributes.items()])
                s += f", attributes: {attributes_str}"
            return s

        tracer_exporter = LogSpanExporter(formatter=custom_formatter, log_out=self.log_out)
        processor = SimpleSpanProcessor(tracer_exporter)
        provider.add_span_processor(span_processor=processor)

    def _add_otel_trace_processor(self, provider: TracerProvider) -> None:
        otel_endpoint = self.get_otel_endpoint()
        if not otel_endpoint:
            return None
        from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

        logger.debug(f"adding otel exported at {otel_endpoint}")
        trace_exporter = OTLPSpanExporter()
        processor = SimpleSpanProcessor(trace_exporter)
        provider.add_span_processor(processor)

    def init_trace_provider(self, resource: Resource) -> TracerProvider:
        trace_provider = TracerProvider(resource=resource)
        add_fns: list[AddTraceCallable] = [
            self._add_otel_trace_processor,
            self._add_console_trace_processor,
        ]
        for add_fn in add_fns:
            add_fn(provider=trace_provider)
        return trace_provider

    def get_tracer(self) -> Tracer:
        return trace.get_tracer(self.service_name)



from .chunker import Chunker, ChunkerConfig
from .embedder import Embedder, EmbedderConfig
from .filter import Filterer, FiltererConfig
from .partitioner import Partitioner, PartitionerConfig
from .uncompress import UncompressConfig, Uncompressor

__all__ = [
    "Chunker",
    "ChunkerConfig",
    "Embedder",
    "EmbedderConfig",
    "Filterer",
    "FiltererConfig",
    "Partitioner",
    "PartitionerConfig",
    "Uncompressor",
    "UncompressConfig",
]



from pydantic import BaseModel, Field


class Entity(BaseModel):
    type: str
    entity: str


class EntityRelationship(BaseModel):
    to: str
    from_: str = Field(..., alias="from")
    relationship: str


class EntitiesData(BaseModel):
    items: list[Entity] = Field(default_factory=list)
    relationships: list[EntityRelationship] = Field(default_factory=list)



from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.__version__ import __version__ as unstructured_io_ingest_version
from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.duckdb.base import BaseDuckDBUploadStager
from unstructured_ingest.utils.data_prep import get_data_df
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from duckdb import DuckDBPyConnection as MotherDuckConnection
    from pandas import DataFrame

CONNECTOR_TYPE = "motherduck"


class MotherDuckAccessConfig(AccessConfig):
    md_token: str = Field(default=None, description="MotherDuck token")


class MotherDuckConnectionConfig(ConnectionConfig):
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    database: str = Field(
        description="Database name. Name of the MotherDuck database.",
    )
    db_schema: Optional[str] = Field(
        default="main",
        description="Schema name. Schema in the database where the elements table is located.",
    )
    table: Optional[str] = Field(
        default="elements",
        description="Table name. Table name into which the elements data is inserted.",
    )
    access_config: Secret[MotherDuckAccessConfig] = Field(
        default=MotherDuckAccessConfig(), validate_default=True
    )

    @requires_dependencies(["duckdb"], extras="duckdb")
    @contextmanager
    def get_client(self) -> Generator["MotherDuckConnection", None, None]:
        import duckdb

        access_config = self.access_config.get_secret_value()
        with duckdb.connect(
            f"md:?motherduck_token={access_config.md_token}",
            config={
                "custom_user_agent": f"unstructured-io-ingest/{unstructured_io_ingest_version}"
            },
        ) as conn:
            conn.sql(f'USE "{self.database}"')
            yield conn

    @contextmanager
    def get_cursor(self) -> Generator["MotherDuckConnection", None, None]:
        with self.get_client() as client, client.cursor() as cursor:
            yield cursor


class MotherDuckUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class MotherDuckUploadStager(BaseDuckDBUploadStager):
    upload_stager_config: MotherDuckUploadStagerConfig = field(
        default_factory=lambda: MotherDuckUploadStagerConfig()
    )


class MotherDuckUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=50, description="[Not-used] Number of records per batch")


@dataclass
class MotherDuckUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: MotherDuckUploaderConfig
    connection_config: MotherDuckConnectionConfig

    def precheck(self) -> None:
        try:
            with self.connection_config.get_cursor() as cursor:
                cursor.execute("SELECT 1;")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def upload_dataframe(self, df: "DataFrame") -> None:
        logger.debug(f"uploading {len(df)} entries to {self.connection_config.database} ")
        database = self.connection_config.database
        db_schema = self.connection_config.db_schema
        table = self.connection_config.table

        with self.connection_config.get_client() as conn:
            conn.query(f'INSERT INTO "{database}"."{db_schema}"."{table}" BY NAME SELECT * FROM df')

    @requires_dependencies(["pandas"], extras="duckdb")
    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data=data)
        self.upload_dataframe(df=df)

    @requires_dependencies(["pandas"], extras="duckdb")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path)
        self.upload_dataframe(df=df)


motherduck_destination_entry = DestinationRegistryEntry(
    connection_config=MotherDuckConnectionConfig,
    uploader=MotherDuckUploader,
    uploader_config=MotherDuckUploaderConfig,
    upload_stager=MotherDuckUploadStager,
    upload_stager_config=MotherDuckUploadStagerConfig,
)



import os
import ssl

import certifi


def ssl_context_with_optional_ca_override():
    """
    # https://www.python-httpx.org/advanced/ssl/#working-with-ssl_cert_file-and-ssl_cert_dir
    # We choose REQUESTS_CA_BUNDLE because that works with many other Python packages.
    """
    return ssl.create_default_context(
        cafile=os.environ.get("REQUESTS_CA_BUNDLE", certifi.where()),
        capath=os.environ.get("REQUESTS_CA_BUNDLE"),
    )



from typing import Any, Dict, Optional

from unstructured_ingest.logger import logger
from unstructured_ingest.processes.utils.logging.sanitizer import DataSanitizer


class LoggingConfig:
    """Configuration for connector logging behavior."""

    def __init__(
        self,
        log_file_paths: bool = False,
        log_document_locations: Optional[bool] = None,
        log_ids: bool = False,
        log_document_ids: Optional[bool] = None,
        log_progress_interval: int = 10,
        sanitize_logs: bool = True,
        show_connection_details: bool = False,
    ):
        # Backward compatibility: if new parameters aren't specified, use old ones
        self.log_file_paths = log_file_paths
        self.log_document_locations = (
            log_document_locations if log_document_locations is not None else log_file_paths
        )

        self.log_ids = log_ids
        self.log_document_ids = log_document_ids if log_document_ids is not None else log_ids

        self.log_progress_interval = log_progress_interval
        self.sanitize_logs = sanitize_logs
        self.show_connection_details = show_connection_details


class ConnectorLoggingMixin:
    """Mixin class providing standardized logging patterns for connectors."""

    def __init__(self, *args, **kwargs):
        """
        Initialize the mixin by setting up logging configuration and data sanitization.

        This method ensures that the mixin provides standardized logging patterns for connectors.
        It initializes:
        - `_logging_config`: Manages logging behavior and settings.
        - `_sanitizer`: Handles sanitization of sensitive data in logs.

        Args:
            *args: Positional arguments passed to the parent class.
            **kwargs: Keyword arguments passed to the parent class.
        """
        super().__init__(*args, **kwargs)
        self._logging_config = LoggingConfig()
        self._sanitizer = DataSanitizer()

    def set_logging_config(self, config: LoggingConfig):
        """Set the logging configuration for this connector."""
        self._logging_config = config

    def _should_sanitize(self) -> bool:
        """Check if log sanitization is enabled."""
        return self._logging_config.sanitize_logs

    def log_operation_start(self, operation: str, **kwargs):
        """Log the start of a major operation."""
        logger.info("Starting %s", operation)

        if kwargs:
            if self._should_sanitize():
                sanitized_kwargs = self._sanitizer.sanitize_dict(kwargs)
                logger.debug("%s parameters: %s", operation, sanitized_kwargs)
            else:
                logger.debug("%s parameters: %s", operation, kwargs)

    def log_operation_complete(self, operation: str, count: Optional[int] = None, **kwargs):
        """Log the completion of a major operation."""
        if count is not None:
            logger.info("Completed %s (%s items)", operation, count)
        else:
            logger.info("Completed %s", operation)

        if kwargs:
            if self._should_sanitize():
                sanitized_kwargs = self._sanitizer.sanitize_dict(kwargs)
                logger.debug("%s results: %s", operation, sanitized_kwargs)
            else:
                logger.debug("%s results: %s", operation, kwargs)

    def log_connection_validated(self, connector_type: str, endpoint: Optional[str] = None):
        """Log successful connection validation."""
        if self._logging_config.show_connection_details and endpoint:
            if self._should_sanitize():
                sanitized_endpoint = self._sanitizer.sanitize_url(endpoint)
                logger.debug(
                    "Connection to %s validated successfully: %s",
                    connector_type,
                    sanitized_endpoint,
                )
            else:
                logger.debug(
                    "Connection to %s validated successfully: %s", connector_type, endpoint
                )
        else:
            logger.debug("Connection to %s validated successfully", connector_type)

    def log_connection_failed(
        self, connector_type: str, error: Exception, endpoint: Optional[str] = None
    ):
        """Log connection validation failure."""
        if endpoint:
            if self._should_sanitize():
                sanitized_endpoint = self._sanitizer.sanitize_url(endpoint)
                logger.error(
                    "Failed to validate %s connection to %s: %s",
                    connector_type,
                    sanitized_endpoint,
                    error,
                    exc_info=True,
                )
            else:
                logger.error(
                    "Failed to validate %s connection to %s: %s",
                    connector_type,
                    endpoint,
                    error,
                    exc_info=True,
                )
        else:
            logger.error(
                "Failed to validate %s connection: %s", connector_type, error, exc_info=True
            )

    def log_progress(
        self, current: int, total: int, item_type: str = "items", operation: str = "Processing"
    ):
        """Log progress for long-running operations."""
        if total > 0 and current % self._logging_config.log_progress_interval == 0:
            progress = (current / total) * 100
            logger.info("%s: %s/%s %s (%.1f%%)", operation, current, total, item_type, progress)

    def log_batch_progress(
        self, batch_num: int, total_batches: int, batch_size: int, operation: str = "Processing"
    ):
        """Log progress for batch operations."""
        logger.info("%s batch %s/%s (%s items)", operation, batch_num, total_batches, batch_size)

    def log_document_operation(
        self,
        operation: str,
        document_location: Optional[str] = None,
        document_id: Optional[str] = None,
        content_size: Optional[int] = None,
        **kwargs,
    ):
        """Log document-related operations (universal for all connector types)."""
        if self._logging_config.log_document_locations and document_location:
            if self._should_sanitize():
                sanitized_location = self._sanitizer.sanitize_location(document_location)
                logger.debug("%s: %s", operation, sanitized_location)
            else:
                logger.debug("%s: %s", operation, document_location)
        elif self._logging_config.log_document_ids and document_id:
            if self._should_sanitize():
                sanitized_id = self._sanitizer.sanitize_document_id(document_id)
                logger.debug("%s: %s", operation, sanitized_id)
            else:
                logger.debug("%s: %s", operation, document_id)
        else:
            logger.debug("%s: <document>", operation)

        if content_size is not None:
            kwargs["content_size"] = content_size

        if kwargs:
            if self._should_sanitize():
                sanitized_kwargs = self._sanitizer.sanitize_dict(kwargs)
                logger.debug("%s details: %s", operation, sanitized_kwargs)
            else:
                logger.debug("%s details: %s", operation, kwargs)

    def log_file_operation(
        self,
        operation: str,
        file_path: Optional[str] = None,
        file_id: Optional[str] = None,
        **kwargs,
    ):
        """Log file-related operations (backward compatibility wrapper)."""
        self.log_document_operation(
            operation=operation, document_location=file_path, document_id=file_id, **kwargs
        )

    def log_document_download_start(
        self,
        document_location: Optional[str] = None,
        document_id: Optional[str] = None,
        content_size: Optional[int] = None,
    ):
        """Log the start of a document download/retrieval."""
        logger.info("Starting document download")

        self.log_document_operation(
            "Download",
            document_location=document_location,
            document_id=document_id,
            content_size=content_size,
        )

    def log_document_download_complete(
        self,
        document_location: Optional[str] = None,
        document_id: Optional[str] = None,
        download_path: Optional[str] = None,
        content_size: Optional[int] = None,
        items_retrieved: Optional[int] = None,
    ):
        """Log the completion of a document download/retrieval."""
        logger.info("Document download completed")

        details = {}
        if download_path:
            details["download_path"] = download_path
        if items_retrieved is not None:
            details["items_retrieved"] = items_retrieved

        self.log_document_operation(
            "Download completed",
            document_location=document_location,
            document_id=document_id,
            content_size=content_size,
            **details,
        )

    def log_download_start(
        self,
        file_path: Optional[str] = None,
        file_id: Optional[str] = None,
        file_size: Optional[int] = None,
    ):
        """Log the start of a file download (backward compatibility wrapper)."""
        self.log_document_download_start(
            document_location=file_path, document_id=file_id, content_size=file_size
        )

    def log_download_complete(
        self,
        file_path: Optional[str] = None,
        file_id: Optional[str] = None,
        download_path: Optional[str] = None,
        file_size: Optional[int] = None,
    ):
        """Log the completion of a file download (backward compatibility wrapper)."""
        self.log_document_download_complete(
            document_location=file_path,
            document_id=file_id,
            download_path=download_path,
            content_size=file_size,
        )

    def log_upload_start(
        self,
        file_path: Optional[str] = None,
        destination: Optional[str] = None,
        file_size: Optional[int] = None,
    ):
        """Log the start of a file upload."""
        logger.info("Starting file upload")

        details = {}
        if destination:
            details["destination"] = destination

        self.log_file_operation("Upload", file_path=file_path, **details)

    def log_upload_complete(
        self,
        file_path: Optional[str] = None,
        destination: Optional[str] = None,
        file_id: Optional[str] = None,
        file_size: Optional[int] = None,
    ):
        """Log the completion of a file upload."""
        logger.info("File upload completed")

        details = {}
        if destination:
            details["destination"] = destination
        if file_id:
            details["file_id"] = file_id

        self.log_file_operation("Upload completed", file_path=file_path, **details)

    def log_indexing_start(self, source_type: str, count: Optional[int] = None):
        """Log the start of indexing operation."""
        if count:
            logger.info("Starting indexing of %s (%s items)", source_type, count)
        else:
            logger.info("Starting indexing of %s", source_type)

    def log_indexing_complete(self, source_type: str, count: int):
        """Log the completion of indexing operation."""
        logger.info("Indexing completed: %s %s items indexed", count, source_type)

    def log_info(self, message: str, context: Optional[Dict[str, Any]] = None, **kwargs):
        """Log an info message with optional context and sanitization."""
        logger.info(message)
        self._log_context("Info", context, **kwargs)

    def log_debug(self, message: str, context: Optional[Dict[str, Any]] = None, **kwargs):
        """Log a debug message with optional context and sanitization."""
        logger.debug(message)
        self._log_context("Debug", context, **kwargs)

    def log_warning(self, message: str, context: Optional[Dict[str, Any]] = None, **kwargs):
        """Log a warning message with optional context and sanitization."""
        logger.warning(message)
        self._log_context("Warning", context, **kwargs)

    def log_error(
        self,
        message: str,
        error: Optional[Exception] = None,
        context: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """Log an error message with optional exception, context and sanitization."""
        if error:
            logger.error("%s: %s", message, error, exc_info=True)
        else:
            logger.error(message)
        self._log_context("Error", context, **kwargs)

    def _log_context(self, log_type: str, context: Optional[Dict[str, Any]], **kwargs):
        """Helper method to log context with sanitization."""
        all_context = {}
        if context:
            all_context.update(context)
        if kwargs:
            all_context.update(kwargs)

        if all_context:
            if self._should_sanitize():
                sanitized_context = self._sanitizer.sanitize_dict(all_context)
                logger.debug("%s context: %s", log_type, sanitized_context)
            else:
                logger.debug("%s context: %s", log_type, all_context)

    def log_api_call(self, method: str, endpoint: str, status_code: Optional[int] = None, **kwargs):
        """Log API call details."""
        if self._should_sanitize():
            sanitized_endpoint = self._sanitizer.sanitize_url(endpoint)
            if status_code:
                logger.debug("API call: %s %s -> %s", method, sanitized_endpoint, status_code)
            else:
                logger.debug("API call: %s %s", method, sanitized_endpoint)
        else:
            if status_code:
                logger.debug("API call: %s %s -> %s", method, endpoint, status_code)
            else:
                logger.debug("API call: %s %s", method, endpoint)

        if kwargs:
            if self._should_sanitize():
                sanitized_kwargs = self._sanitizer.sanitize_dict(kwargs)
                logger.debug("API call details: %s", sanitized_kwargs)
            else:
                logger.debug("API call details: %s", kwargs)



import glob
import json
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import Any, Generator

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import FileExistsError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)

CONNECTOR_TYPE = "local"


class LocalAccessConfig(AccessConfig):
    pass


class LocalConnectionConfig(ConnectionConfig):
    access_config: Secret[LocalAccessConfig] = Field(
        default=LocalAccessConfig(), validate_default=True
    )


class LocalIndexerConfig(IndexerConfig):
    input_path: Path = Field(
        description="Path to the location in the local file system that will be processed."
    )
    recursive: bool = Field(
        default=False,
        description="Recursively download files in their respective folders "
        "otherwise stop at the files in provided folder level.",
    )

    @property
    def path(self) -> Path:
        return Path(self.input_path).resolve()


@dataclass
class LocalIndexer(Indexer):
    index_config: LocalIndexerConfig
    connection_config: LocalConnectionConfig = field(
        default_factory=lambda: LocalConnectionConfig()
    )
    connector_type: str = CONNECTOR_TYPE

    def list_files(self) -> list[Path]:
        input_path = self.index_config.path
        if input_path.is_file():
            return [Path(s) for s in glob.glob(f"{self.index_config.path}")]
        files = []
        if self.index_config.recursive:
            files.extend(list(input_path.rglob("*")))
        else:
            files.extend(list(input_path.glob("*")))
        return [f for f in files if f.is_file()]

    def get_file_metadata(self, path: Path) -> FileDataSourceMetadata:
        stats = path.stat()
        try:
            date_modified = str(stats.st_mtime)
        except Exception as e:
            logger.warning(f"Couldn't detect date modified: {e}")
            date_modified = None

        try:
            date_created = str(stats.st_birthtime)
        except Exception as e:
            logger.warning(f"Couldn't detect date created: {e}")
            date_created = None

        try:
            mode = stats.st_mode
            permissions_data = [{"mode": mode}]
        except Exception as e:
            logger.warning(f"Couldn't detect file mode: {e}")
            permissions_data = None

        try:
            filesize_bytes = stats.st_size
        except Exception as e:
            logger.warning(f"Couldn't detect file size: {e}")
            filesize_bytes = None

        return FileDataSourceMetadata(
            date_modified=date_modified,
            date_created=date_created,
            date_processed=str(time()),
            permissions_data=permissions_data,
            record_locator={"path": str(path.resolve())},
            filesize_bytes=filesize_bytes,
        )

    def run(self, **kwargs: Any) -> Generator[FileData, None, None]:
        for file_path in self.list_files():
            source_identifiers = SourceIdentifiers(
                fullpath=str(file_path.resolve()),
                filename=file_path.name,
                rel_path=(
                    str(file_path.resolve()).replace(str(self.index_config.path.resolve()), "")[1:]
                    if not self.index_config.path.is_file()
                    else self.index_config.path.name
                ),
            )
            file_data = FileData(
                identifier=str(file_path.resolve()),
                connector_type=CONNECTOR_TYPE,
                source_identifiers=source_identifiers,
                metadata=self.get_file_metadata(path=file_path),
                display_name=source_identifiers.fullpath,
            )
            yield file_data


class LocalDownloaderConfig(DownloaderConfig):
    pass


@dataclass
class LocalDownloader(Downloader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: LocalConnectionConfig = field(default_factory=LocalConnectionConfig)
    download_config: LocalDownloaderConfig = field(default_factory=LocalDownloaderConfig)

    def get_download_path(self, file_data: FileData) -> Path:
        return Path(file_data.source_identifiers.fullpath)

    def run(self, file_data: FileData, **kwargs: Any) -> DownloadResponse:
        return DownloadResponse(
            file_data=file_data, path=Path(file_data.source_identifiers.fullpath)
        )


class LocalUploaderConfig(UploaderConfig):
    output_dir: str = Field(
        default="structured-output", description="Local path to write partitioned output to"
    )

    @property
    def output_path(self) -> Path:
        return Path(self.output_dir).resolve()

    def __post_init__(self):
        if self.output_path.exists() and self.output_path.is_file():
            raise FileExistsError(f"output path {self.output_path} already exists as a file")


@dataclass
class LocalUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: LocalUploaderConfig = field(default_factory=LocalUploaderConfig)
    connection_config: LocalConnectionConfig = field(
        default_factory=lambda: LocalConnectionConfig()
    )

    def is_async(self) -> bool:
        return False

    def get_destination_path(self, file_data: FileData) -> Path:
        if source_identifiers := file_data.source_identifiers:
            rel_path = (
                source_identifiers.relative_path[1:]
                if source_identifiers.relative_path.startswith("/")
                else source_identifiers.relative_path
            )
            new_path = self.upload_config.output_path / Path(rel_path)
            final_path = str(new_path).replace(
                source_identifiers.filename, f"{source_identifiers.filename}.json"
            )
        else:
            final_path = self.upload_config.output_path / Path(f"{file_data.identifier}.json")
        final_path = Path(final_path)
        final_path.parent.mkdir(parents=True, exist_ok=True)
        return final_path

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        final_path = self.get_destination_path(file_data=file_data)
        with final_path.open("w") as f:
            json.dump(data, f)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        final_path = self.get_destination_path(file_data=file_data)
        logger.debug(f"copying file from {path} to {final_path}")
        shutil.copy(src=str(path), dst=str(final_path))


local_source_entry = SourceRegistryEntry(
    indexer=LocalIndexer,
    indexer_config=LocalIndexerConfig,
    downloader=LocalDownloader,
    downloader_config=LocalDownloaderConfig,
    connection_config=LocalConnectionConfig,
)

local_destination_entry = DestinationRegistryEntry(
    uploader=LocalUploader,
    uploader_config=LocalUploaderConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



# https://developers.notion.com/reference/user
from dataclasses import dataclass, field
from typing import Optional

from htmlBuilder.attributes import Href
from htmlBuilder.tags import A, Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import (
    FromJSONMixin,
    GetHTMLMixin,
)


@dataclass
class PartialUser(FromJSONMixin):
    id: str
    object: str = "user"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(id=data["id"])


@dataclass
class User(FromJSONMixin, GetHTMLMixin):
    object: dict
    id: str
    type: Optional[str] = None
    name: Optional[str] = None
    avatar_url: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_text(self) -> Optional[str]:
        text = self.name
        if self.avatar_url:
            text = f"[{text}]({self.avatar_url})"
        return text

    def get_html(self) -> Optional[HtmlTag]:
        if self.avatar_url and self.name:
            return A([Href(self.avatar_url)], self.name)
        elif self.name:
            return Div([], self.name)
        else:
            return Div([], "")


@dataclass
class People(User):
    person: dict = field(default_factory=dict)


@dataclass
class Bots(FromJSONMixin, GetHTMLMixin):
    object: dict
    id: str
    bot: dict
    owner: dict
    type: str
    workspace_name: str
    name: Optional[str] = None
    avatar_url: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_text(self) -> Optional[str]:
        text = self.name
        if self.avatar_url:
            text = f"[{text}]({self.avatar_url})"
        return text

    def get_html(self) -> Optional[HtmlTag]:
        if self.avatar_url and self.name:
            return A([Href(self.avatar_url)], self.name)
        elif self.name:
            return Div([], self.name)
        else:
            return Div([], "")



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from pydantic import Field, SecretStr

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.error import (
    ProviderError,
    QuotaError,
    RateLimitError,
    UserAuthError,
    UserError,
    is_internal_error,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.tls import ssl_context_with_optional_ca_override

if TYPE_CHECKING:
    from openai import AsyncOpenAI, OpenAI


class OpenAIEmbeddingConfig(EmbeddingConfig):
    api_key: SecretStr = Field(description="API key for OpenAI")
    embedder_model_name: str = Field(
        default="text-embedding-ada-002", alias="model_name", description="OpenAI model name"
    )
    base_url: Optional[str] = Field(default=None, description="optional override for the base url")

    @requires_dependencies(["openai"], extras="openai")
    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        # https://platform.openai.com/docs/guides/error-codes/api-errors
        from openai import APIStatusError

        if not isinstance(e, APIStatusError):
            logger.error(f"unhandled exception from openai: {e}", exc_info=True)
            raise e
        error_code = e.code
        if 400 <= e.status_code < 500:
            # user error
            if e.status_code == 401:
                return UserAuthError(e.message)
            if e.status_code == 429:
                # 429 indicates rate limit exceeded and quote exceeded
                if error_code == "insufficient_quota":
                    return QuotaError(e.message)
                else:
                    return RateLimitError(e.message)
            return UserError(e.message)
        if e.status_code >= 500:
            return ProviderError(e.message)
        logger.error(f"unhandled exception from openai: {e}", exc_info=True)
        return e

    @requires_dependencies(["openai"], extras="openai")
    def get_models(self) -> Optional[list[str]]:
        # In case the list model endpoint isn't exposed, don't break
        from openai import APIStatusError

        client = self.get_client()
        try:
            models = [m.id for m in list(client.models.list())]
            return models
        except APIStatusError as e:
            if e.status_code == 404:
                return None
        except Exception as e:
            raise self.wrap_error(e=e)

    def run_precheck(self) -> None:
        try:
            models = self.get_models()
            if models is None:
                return
            if self.embedder_model_name not in models:
                raise UserError(
                    "model '{}' not found: {}".format(self.embedder_model_name, ", ".join(models))
                )
        except Exception as e:
            raise self.wrap_error(e=e)

    @requires_dependencies(["openai"], extras="openai")
    def get_client(self) -> "OpenAI":
        from openai import DefaultHttpxClient, OpenAI

        client = DefaultHttpxClient(verify=ssl_context_with_optional_ca_override())
        return OpenAI(
            api_key=self.api_key.get_secret_value(), http_client=client, base_url=self.base_url
        )

    @requires_dependencies(["openai"], extras="openai")
    def get_async_client(self) -> "AsyncOpenAI":
        from openai import AsyncOpenAI, DefaultAsyncHttpxClient

        client = DefaultAsyncHttpxClient(verify=ssl_context_with_optional_ca_override())
        return AsyncOpenAI(
            api_key=self.api_key.get_secret_value(), http_client=client, base_url=self.base_url
        )


@dataclass
class OpenAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: OpenAIEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "OpenAI":
        return self.config.get_client()

    def embed_batch(self, client: "OpenAI", batch: list[str]) -> list[list[float]]:
        response = client.embeddings.create(input=batch, model=self.config.embedder_model_name)
        return [data.embedding for data in response.data]


@dataclass
class AsyncOpenAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: OpenAIEmbeddingConfig

    def precheck(self):
        self.config.run_precheck()

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "AsyncOpenAI":
        return self.config.get_async_client()

    async def embed_batch(self, client: "AsyncOpenAI", batch: list[str]) -> list[list[float]]:
        response = await client.embeddings.create(
            input=batch, model=self.config.embedder_model_name
        )
        return [data.embedding for data in response.data]



from dataclasses import dataclass
from typing import TYPE_CHECKING

from pydantic import Field

from unstructured_ingest.embed.openai import (
    AsyncOpenAIEmbeddingEncoder,
    OpenAIEmbeddingConfig,
    OpenAIEmbeddingEncoder,
)
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.tls import ssl_context_with_optional_ca_override

if TYPE_CHECKING:
    from openai import AsyncAzureOpenAI, AzureOpenAI


class AzureOpenAIEmbeddingConfig(OpenAIEmbeddingConfig):
    api_version: str = Field(description="Azure API version", default="2024-06-01")
    azure_endpoint: str = Field(description="Azure endpoint")
    embedder_model_name: str = Field(
        default="text-embedding-ada-002", alias="model_name", description="Azure OpenAI model name"
    )

    @requires_dependencies(["openai"], extras="openai")
    def get_client(self) -> "AzureOpenAI":
        from openai import AzureOpenAI, DefaultHttpxClient

        client = DefaultHttpxClient(verify=ssl_context_with_optional_ca_override())
        return AzureOpenAI(
            http_client=client,
            api_key=self.api_key.get_secret_value(),
            api_version=self.api_version,
            azure_endpoint=self.azure_endpoint,
        )

    @requires_dependencies(["openai"], extras="openai")
    def get_async_client(self) -> "AsyncAzureOpenAI":
        from openai import AsyncAzureOpenAI, DefaultAsyncHttpxClient

        client = DefaultAsyncHttpxClient(verify=ssl_context_with_optional_ca_override())
        return AsyncAzureOpenAI(
            http_client=client,
            api_key=self.api_key.get_secret_value(),
            api_version=self.api_version,
            azure_endpoint=self.azure_endpoint,
        )


@dataclass
class AzureOpenAIEmbeddingEncoder(OpenAIEmbeddingEncoder):
    config: AzureOpenAIEmbeddingConfig

    def get_client(self) -> "AzureOpenAI":
        return self.config.get_client()


@dataclass
class AsyncAzureOpenAIEmbeddingEncoder(AsyncOpenAIEmbeddingEncoder):
    config: AzureOpenAIEmbeddingConfig

    def get_client(self) -> "AsyncAzureOpenAI":
        return self.config.get_async_client()



from abc import ABC
from functools import wraps
from typing import Optional


class UnstructuredIngestError(Exception, ABC):
    error_string: str
    status_code: Optional[int] = None

    @classmethod
    def wrap(cls, f):
        """
        Provides a wrapper for a function that catches any exception and
        re-raises it as the customer error. If the exception itself is already an instance
        of the custom error, re-raises original error.
        """

        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as error:
                if not isinstance(error, cls) and not issubclass(type(error), cls):
                    raise cls(cls.error_string.format(str(error))) from error
                raise

        return wrapper


class ConnectionError(UnstructuredIngestError):
    error_string = "Connection error: {}"
    status_code: Optional[int] = 400


class SourceConnectionError(ConnectionError):
    error_string = "Error in getting data from upstream data source: {}"
    status_code: Optional[int] = 400


class SourceConnectionNetworkError(SourceConnectionError):
    error_string = "Error in connecting to upstream data source: {}"
    status_code: Optional[int] = 400


class DestinationConnectionError(ConnectionError):
    error_string = "Error in connecting to downstream data source: {}"
    status_code: Optional[int] = 400


class EmbeddingEncoderConnectionError(ConnectionError):
    error_string = "Error in connecting to the embedding model provider: {}"
    status_code: Optional[int] = 400


class UserError(UnstructuredIngestError):
    error_string = "User error: {}"
    status_code: Optional[int] = 401


class UserAuthError(UserError):
    error_string = "User authentication error: {}"
    status_code: Optional[int] = 401


class RateLimitError(UserError):
    error_string = "Rate limit error: {}"
    status_code: Optional[int] = 429


class NotFoundError(UnstructuredIngestError):
    error_string = "Not found error: {}"
    status_code: Optional[int] = 404


class TimeoutError(UnstructuredIngestError):
    error_string = "Timeout error: {}"
    status_code: Optional[int] = 408


class ResponseError(UnstructuredIngestError):
    error_string = "Response error: {}"
    status_code: Optional[int] = 400


class WriteError(UnstructuredIngestError):
    error_string = "Error in writing to downstream data source: {}"
    status_code: Optional[int] = 400


class ProviderError(UnstructuredIngestError):
    error_string = "Provider error: {}"
    status_code: Optional[int] = 500


class ValueError(UnstructuredIngestError):
    error_string = "Value error: {}"


class PartitionError(UnstructuredIngestError):
    error_string = "Error in partitioning content: {}"


class QuotaError(UserError):
    error_string = "Quota error: {}"


class MissingCategoryError(UnstructuredIngestError):
    error_string = "Missing category error: {}"


class ValidationError(UnstructuredIngestError):
    error_string = "Validation error: {}"


class KeyError(UnstructuredIngestError):
    error_string = "Key error: {}"


class FileExistsError(UnstructuredIngestError):
    error_string = "File exists error: {}"


class TypeError(UnstructuredIngestError):
    error_string = "Type error: {}"


class IcebergCommitFailedException(UnstructuredIngestError):
    error_string = "Failed to commit changes to the iceberg table"


recognized_errors = [
    UserError,
    UserAuthError,
    RateLimitError,
    QuotaError,
    ProviderError,
    NotFoundError,
    TypeError,
    ValueError,
    FileExistsError,
    TimeoutError,
    KeyError,
    ResponseError,
    ValidationError,
    PartitionError,
    WriteError,
    ConnectionError,
    SourceConnectionError,
    SourceConnectionNetworkError,
    DestinationConnectionError,
    EmbeddingEncoderConnectionError,
]


def is_internal_error(e: Exception) -> bool:
    return any(isinstance(e, recognized_error) for recognized_error in recognized_errors)



import logging

logger = logging.getLogger("unstructured_ingest")


def remove_root_handlers(logger: logging.Logger) -> None:
    # NOTE(robinson): in some environments such as Google Colab, there is a root handler
    # that doesn't not mask secrets, meaning sensitive info such as api keys appear in logs.
    # Removing these when they exist prevents this behavior
    if logger.root.hasHandlers():
        for handler in logger.root.handlers:
            logger.root.removeHandler(handler)


def ingest_log_streaming_init(level: int) -> None:
    handler = logging.StreamHandler()
    handler.name = "ingest_log_handler"
    formatter = logging.Formatter("%(asctime)s %(processName)-10s %(levelname)-8s %(message)s")
    handler.setFormatter(formatter)

    # Only want to add the handler once
    if "ingest_log_handler" not in [h.name for h in logger.handlers]:
        logger.addHandler(handler)

    remove_root_handlers(logger)
    logger.setLevel(level)


def make_default_logger(level: int) -> logging.Logger:
    """Return a custom logger."""
    logger = logging.getLogger("unstructured_ingest")
    handler = logging.StreamHandler()
    handler.name = "ingest_log_handler"
    formatter = logging.Formatter("%(asctime)s %(processName)-10s %(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)
    remove_root_handlers(logger)
    return logger



import json
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import (
    DestinationConnectionError,
    NotFoundError,
    UnstructuredIngestError,
    UserError,
)
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
    VectorDBUploader,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.utils import ndjson
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.data_prep import (
    flatten_dict,
    generator_batching_wbytes,
    get_enhanced_element_id,
    get_json_data,
    write_data,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from pinecone import Index as PineconeIndex
    from pinecone import Pinecone


CONNECTOR_TYPE = "pinecone"
MAX_PAYLOAD_SIZE = 2 * 1024 * 1024  # 2MB
MAX_POOL_THREADS = 100
MAX_METADATA_BYTES = 40960  # 40KB https://docs.pinecone.io/reference/quotas-and-limits#hard-limits
MAX_QUERY_RESULTS = 10000


class PineconeAccessConfig(AccessConfig):
    pinecone_api_key: Optional[str] = Field(
        default=None, description="API key for Pinecone.", alias="api_key"
    )


class PineconeConnectionConfig(ConnectionConfig):
    index_name: Optional[str] = Field(description="Name of the index to connect to.", default=None)
    access_config: Secret[PineconeAccessConfig] = Field(
        default=PineconeAccessConfig(), validate_default=True
    )

    @requires_dependencies(["pinecone"], extras="pinecone")
    def get_client(self, **index_kwargs) -> "Pinecone":
        from pinecone import Pinecone

        from unstructured_ingest import __version__ as unstructured_version

        return Pinecone(
            api_key=self.access_config.get_secret_value().pinecone_api_key,
            source_tag=f"unstructured_ingest=={unstructured_version}",
        )

    def get_index(self, **index_kwargs) -> "PineconeIndex":
        pc = self.get_client()

        index = pc.Index(name=self.index_name, **index_kwargs)

        logger.debug(f"connected to index: {pc.describe_index(self.index_name)}")
        return index


ALLOWED_FIELDS = (
    "element_id",
    "text",
    "parent_id",
    "category_depth",
    "emphasized_text_tags",
    "emphasized_text_contents",
    "coordinates",
    "last_modified",
    "page_number",
    "filename",
    "is_continuation",
    "link_urls",
    "link_texts",
    "text_as_html",
    "entities",
)


class PineconeUploadStagerConfig(UploadStagerConfig):
    metadata_fields: list[str] = Field(
        default=list(ALLOWED_FIELDS),
        description=(
            "which metadata from the source element to map to the payload metadata being sent to "
            "Pinecone."
        ),
    )


class PineconeUploaderConfig(UploaderConfig):
    batch_size: Optional[int] = Field(
        default=None,
        description="Optional number of records per batch. Will otherwise limit by size.",
    )
    pool_threads: Optional[int] = Field(
        default=1, description="Optional limit on number of threads to use for upload"
    )
    namespace: Optional[str] = Field(
        default=None,
        description="The namespace to write to. If not specified, the default namespace is used",
    )
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )


@dataclass
class PineconeUploadStager(UploadStager):
    upload_stager_config: PineconeUploadStagerConfig = field(
        default_factory=lambda: PineconeUploadStagerConfig()
    )

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        embeddings = element_dict.pop("embeddings", None)
        metadata: dict[str, Any] = element_dict.pop("metadata", {})
        data_source = metadata.pop("data_source", {})
        coordinates = metadata.pop("coordinates", {})
        pinecone_metadata = {}
        for possible_meta in [element_dict, metadata, data_source, coordinates]:
            pinecone_metadata.update(
                {
                    k: v
                    for k, v in possible_meta.items()
                    if k in self.upload_stager_config.metadata_fields
                }
            )

        metadata = flatten_dict(
            pinecone_metadata,
            separator="-",
            flatten_lists=True,
            remove_none=True,
        )
        metadata_size_bytes = len(json.dumps(metadata).encode())
        if metadata_size_bytes > MAX_METADATA_BYTES:
            logger.info(
                f"Metadata size is {metadata_size_bytes} bytes, which exceeds the limit of"
                f" {MAX_METADATA_BYTES} bytes per vector. Dropping the metadata."
            )
            metadata = {}

        metadata[RECORD_ID_LABEL] = file_data.identifier

        # To support more optimal deletes, a prefix is suggested for each record:
        # https://docs.pinecone.io/guides/data/manage-rag-documents#delete-all-records-for-a-parent-document
        return {
            "id": f"{file_data.identifier}#{get_enhanced_element_id(element_dict=element_dict, file_data=file_data)}",  # noqa:E501
            "values": embeddings,
            "metadata": metadata,
        }

    def stream_update(self, input_file: Path, output_file: Path, file_data: FileData) -> None:
        with input_file.open() as in_f:
            reader = ndjson.reader(in_f)
            with output_file.open("w") as out_f:
                writer = ndjson.writer(out_f)
                for element in reader:
                    if "embeddings" not in element:
                        continue
                    conformed_element = self.conform_dict(element_dict=element, file_data=file_data)
                    writer.write(row=conformed_element)
                    writer.f.flush()

    def process_whole(self, input_file: Path, output_file: Path, file_data: FileData) -> None:
        elements_contents = get_json_data(path=input_file)

        conformed_elements = [
            self.conform_dict(element_dict=element, file_data=file_data)
            for element in elements_contents
            if "embeddings" in element
        ]
        write_data(path=output_file, data=conformed_elements)


@dataclass
class PineconeUploader(VectorDBUploader):
    upload_config: PineconeUploaderConfig
    connection_config: PineconeConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def init(self, **kwargs: Any) -> None:
        self.create_destination(**kwargs)

    def index_exists(self, index_name: Optional[str]) -> bool:
        from pinecone.exceptions import NotFoundException

        index_name = index_name or self.connection_config.index_name
        pc = self.connection_config.get_client()
        try:
            pc.describe_index(index_name)
            return True
        except NotFoundException:
            return False
        except Exception as e:
            logger.error(f"failed to check if pinecone index exists : {e}")
            raise DestinationConnectionError(f"failed to check if pinecone index exists : {e}")

    def precheck(self):
        try:
            # just a connection check here. not an actual index_exists check
            self.index_exists("just-checking-our-connection")

            if self.connection_config.index_name and not self.index_exists(
                self.connection_config.index_name
            ):
                raise NotFoundError(f"index {self.connection_config.index_name} does not exist")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def format_destination_name(self, destination_name: str) -> str:
        # Pinecone naming requirements:
        # can only contain lowercase letters, numbers, and hyphens
        # must be 45 characters or less
        formatted = re.sub(r"[^a-z0-9]", "-", destination_name.lower())
        return formatted

    def create_destination(
        self,
        vector_length: int,
        destination_name: str = "unstructuredautocreated",
        destination_type: Literal["pod", "serverless"] = "serverless",
        serverless_cloud: str = "aws",
        serverless_region: str = "us-east-1",
        pod_environment: str = "us-east1-gcp",
        pod_type: str = "p1.x1",
        pod_count: int = 1,
        **kwargs: Any,
    ) -> bool:
        from pinecone import PodSpec, ServerlessSpec

        index_name = self.connection_config.index_name or destination_name
        index_name = self.format_destination_name(index_name)
        self.connection_config.index_name = index_name

        if not self.index_exists(index_name):
            logger.info(f"creating pinecone index {index_name}")

            pc = self.connection_config.get_client()
            if destination_type == "serverless":
                pc.create_index(
                    name=index_name,
                    dimension=vector_length,
                    spec=ServerlessSpec(cloud=serverless_cloud, region=serverless_region),
                )

                return True

            elif destination_type == "pod":
                pc.create_index(
                    name=destination_name,
                    dimension=vector_length,
                    spec=PodSpec(environment=pod_environment, pod_type=pod_type, pods=pod_count),
                )

                return True

            else:
                raise ValueError(f"unexpected destination type: {destination_type}")

        else:
            logger.debug(f"index {index_name} already exists, skipping creation")
            return False

    def pod_delete_by_record_id(self, file_data: FileData) -> None:
        logger.debug(
            f"deleting any content with metadata "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from pinecone pod index"
        )
        index = self.connection_config.get_index(pool_threads=MAX_POOL_THREADS)
        delete_kwargs = {
            "filter": {self.upload_config.record_id_key: {"$eq": file_data.identifier}}
        }

        if namespace := self.upload_config.namespace:
            delete_kwargs["namespace"] = namespace
            try:
                index.delete(**delete_kwargs)
            except UserError as e:
                logger.error(f"failed to delete batch of ids: {delete_kwargs} {e}")

        logger.debug(
            f"deleted any content with metadata "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from pinecone index: {delete_kwargs}"
        )

    def serverless_delete_by_record_id(self, file_data: FileData) -> None:
        logger.debug(
            f"deleting any content with metadata "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from pinecone serverless index"
        )
        index = self.connection_config.get_index(pool_threads=MAX_POOL_THREADS)
        list_kwargs = {"prefix": f"{file_data.identifier}#"}
        deleted_ids = 0
        if namespace := self.upload_config.namespace:
            list_kwargs["namespace"] = namespace

        for ids in index.list(**list_kwargs):
            deleted_ids += len(ids)
            delete_kwargs = {"ids": ids}

            if namespace := self.upload_config.namespace:
                delete_kwargs["namespace"] = namespace

            try:
                index.delete(**delete_kwargs)
            except UserError as e:
                logger.error(f"failed to delete batch of ids: {delete_kwargs} {e}")

        logger.info(
            f"deleted {deleted_ids} records with metadata "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from pinecone index"
        )

    @requires_dependencies(["pinecone"], extras="pinecone")
    def upsert_batches_async(self, elements_dict: list[dict]):
        from pinecone.exceptions import PineconeApiException

        chunks = list(
            generator_batching_wbytes(
                iterable=elements_dict,
                batch_size_limit_bytes=MAX_PAYLOAD_SIZE - 100,
                max_batch_size=self.upload_config.batch_size,
            )
        )
        logger.info(f"split doc with {len(elements_dict)} elements into {len(chunks)} batches")

        max_pool_threads = min(len(chunks), MAX_POOL_THREADS)
        if self.upload_config.pool_threads:
            pool_threads = min(self.upload_config.pool_threads, max_pool_threads)
        else:
            pool_threads = max_pool_threads
        index = self.connection_config.get_index(pool_threads=pool_threads)
        with index:
            upsert_kwargs = [{"vectors": chunk, "async_req": True} for chunk in chunks]
            if namespace := self.upload_config.namespace:
                for kwargs in upsert_kwargs:
                    kwargs["namespace"] = namespace
            async_results = [index.upsert(**kwarg) for kwarg in upsert_kwargs]
            # Wait for and retrieve responses (this raises in case of error)
            try:
                results = [async_result.get() for async_result in async_results]
            except PineconeApiException as api_error:
                raise UnstructuredIngestError(f"http error: {api_error}") from api_error
            logger.debug(f"results: {results}")

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        logger.info(
            f"writing a total of {len(data)} elements via"
            f" document batches to destination"
            f" index named {self.connection_config.index_name}"
        )
        # Determine if serverless or pod based index
        pinecone_client = self.connection_config.get_client()

        if not self.connection_config.index_name:
            raise ValueError("No index name specified")

        index_description = pinecone_client.describe_index(name=self.connection_config.index_name)
        if "serverless" in index_description.get("spec"):
            self.serverless_delete_by_record_id(file_data=file_data)
        elif "pod" in index_description.get("spec"):
            self.pod_delete_by_record_id(file_data=file_data)
        else:
            raise ValueError(f"unexpected spec type in index description: {index_description}")
        self.upsert_batches_async(elements_dict=data)


pinecone_destination_entry = DestinationRegistryEntry(
    connection_config=PineconeConnectionConfig,
    uploader=PineconeUploader,
    uploader_config=PineconeUploaderConfig,
    upload_stager=PineconeUploadStager,
    upload_stager_config=PineconeUploadStagerConfig,
)



# https://developers.notion.com/reference/property-object#relation
from dataclasses import dataclass
from typing import Optional
from urllib.parse import unquote

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import (
    DBCellBase,
    DBPropertyBase,
    FromJSONMixin,
)


@dataclass
class DualProperty(FromJSONMixin):
    synced_property_id: str
    synced_property_name: str

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class RelationProp(FromJSONMixin):
    database_id: str
    type: str
    dual_property: DualProperty

    @classmethod
    def from_dict(cls, data: dict):
        t = data.get("type")
        if t == "dual_property":
            dual_property = DualProperty.from_dict(data.pop(t))
        else:
            raise ValueError(f"{t} type not recognized")

        return cls(dual_property=dual_property, **data)


@dataclass
class Relation(DBPropertyBase):
    id: str
    name: str
    relation: RelationProp
    type: str = "relation"
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(relation=RelationProp.from_dict(data.pop("relation")), **data)


@dataclass
class RelationCell(DBCellBase):
    id: str
    has_more: bool
    relation: list
    type: str = "relation"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def get_html(self) -> Optional[HtmlTag]:
        return Div([], unquote(self.id))



import json
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncGenerator, Generator, Optional

from pydantic import Field, Secret, model_validator

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError, ResponseError, ValueError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.utils.data_prep import batch_generator
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from redis.asyncio import Redis

import asyncio

CONNECTOR_TYPE = "redis"
SERVER_API_VERSION = "1"


class RedisAccessConfig(AccessConfig):
    uri: Optional[str] = Field(
        default=None, description="If not anonymous, use this uri, if specified."
    )
    password: Optional[str] = Field(
        default=None,
        description="Password used to connect to database if uri is "
        "not specified and connection is not anonymous.",
    )


class RedisConnectionConfig(ConnectionConfig):
    access_config: Secret[RedisAccessConfig] = Field(
        default=RedisAccessConfig(), validate_default=True
    )
    host: Optional[str] = Field(
        default=None,
        description="Hostname or IP address of a Redis instance to connect to "
        "if uri is not specified.",
    )
    database: int = Field(default=0, description="Database index to connect to.")
    port: Optional[int] = Field(
        default=6379, description="Port used to connect to database if uri is not specified."
    )
    username: Optional[str] = Field(
        default=None, description="Username used to connect to database if uri is not specified."
    )
    ssl: Optional[bool] = Field(
        default=True,
        description="Whether the connection should use SSL encryption if uri is not specified.",
    )
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @model_validator(mode="after")
    def validate_host_or_url(self) -> "RedisConnectionConfig":
        if not self.access_config.get_secret_value().uri:
            if not self.host:
                raise ValueError("Please pass a hostname either directly or through uri")
            if self.port is None:
                raise ValueError("Since URI is not specified, port cannot be None")
            if self.ssl is None:
                raise ValueError("Since URI is not specified, ssl cannot be None")
        return self

    @requires_dependencies(["redis"], extras="redis")
    @asynccontextmanager
    async def create_async_client(self) -> AsyncGenerator["Redis", None]:
        from redis.asyncio import Redis, from_url

        access_config = self.access_config.get_secret_value()

        if access_config.uri:
            async with from_url(access_config.uri) as client:
                yield client
        else:
            options = {
                "host": self.host,
                "port": self.port,
                "db": self.database,
                "ssl": self.ssl,
                "username": self.username,
            }

            if access_config.password:
                options["password"] = access_config.password
            async with Redis(**options) as client:
                yield client

    @requires_dependencies(["redis"], extras="redis")
    @contextmanager
    def create_client(self) -> Generator["Redis", None, None]:
        from redis import Redis, from_url

        access_config = self.access_config.get_secret_value()

        options = {
            "host": self.host,
            "port": self.port,
            "db": self.database,
            "ssl": self.ssl,
            "username": self.username,
        }

        if access_config.password:
            options["password"] = access_config.password

        if access_config.uri:
            with from_url(access_config.uri) as client:
                yield client
        else:
            with Redis(**options) as client:
                yield client


class RedisUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=100, description="Number of records per batch")
    key_prefix: str = Field(default="", description="Prefix for Redis keys")


def _form_redis_pipeline_error_message(error: str) -> str:
    """
    Form a user-friendly error message for Redis pipeline errors.
    The error message has `$` character at the beginning and `) of pipeline` at the end.
    Everything between these two strings is the value an should be removed.
    """
    start = error.find("$")
    end = error.find(") of pipeline")
    if start != -1 and end != -1:
        return error[: start + 1] + "<value>" + error[end:]
    else:
        return error


@dataclass
class RedisUploader(Uploader):
    upload_config: RedisUploaderConfig
    connection_config: RedisConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    def is_async(self) -> bool:
        return True

    def precheck(self) -> None:
        try:
            with self.connection_config.create_client() as client:
                client.ping()
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    async def run_data_async(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        first_element = data[0]
        redis_stack = await self._check_redis_stack(first_element)
        logger.info(
            f"writing {len(data)} objects to destination asynchronously, "
            f"db, {self.connection_config.database}, "
            f"at {self.connection_config.host}",
        )

        batches = list(batch_generator(data, batch_size=self.upload_config.batch_size))
        await asyncio.gather(*[self._write_batch(batch, redis_stack) for batch in batches])

    async def _write_batch(self, batch: list[dict], redis_stack: bool) -> None:
        async with (
            self.connection_config.create_async_client() as async_client,
            async_client.pipeline(transaction=True) as pipe,
        ):
            for element in batch:
                key_with_prefix = f"{self.upload_config.key_prefix}{element['element_id']}"
                if redis_stack:
                    pipe.json().set(key_with_prefix, "$", element)
                else:
                    pipe.set(key_with_prefix, json.dumps(element))
            await pipe.execute()

    @requires_dependencies(["redis"], extras="redis")
    async def _check_redis_stack(self, element: dict) -> bool:
        from redis import exceptions as redis_exceptions

        redis_stack = True
        async with (
            self.connection_config.create_async_client() as async_client,
            async_client.pipeline(transaction=True) as pipe,
        ):
            key_with_prefix = f"{self.upload_config.key_prefix}{element['element_id']}"
            try:
                # Redis with stack extension supports JSON type
                await pipe.json().set(key_with_prefix, "$", element).execute()
            except redis_exceptions.ResponseError as e:
                message = _form_redis_pipeline_error_message(str(e))
                if "unknown command `JSON.SET`" in message:
                    # if this error occurs, Redis server doesn't support JSON type,
                    # so save as string type instead
                    await pipe.set(key_with_prefix, json.dumps(element)).execute()
                    redis_stack = False
                else:
                    raise ResponseError(message) from e
        return redis_stack


redis_destination_entry = DestinationRegistryEntry(
    connection_config=RedisConnectionConfig,
    uploader=RedisUploader,
    uploader_config=RedisUploaderConfig,
)



import os
import sys
import tarfile
import zipfile
from pathlib import Path
from typing import Optional

from unstructured_ingest.logger import logger
from unstructured_ingest.utils.filesystem import mkdir_concurrent_safe

ZIP_FILE_EXT = [".zip"]
TAR_FILE_EXT = [".tar", ".tar.gz", ".tgz"]


def uncompress_file(filename: str, path: Optional[str] = None) -> str:
    """
    Takes in a compressed zip or tar file and decompresses it
    """
    # Create path if it doesn't already exist
    if path:
        mkdir_concurrent_safe(Path(path))

    if any(filename.endswith(ext) for ext in ZIP_FILE_EXT):
        return uncompress_zip_file(zip_filename=filename, path=path)
    elif any(filename.endswith(ext) for ext in TAR_FILE_EXT):
        return uncompress_tar_file(tar_filename=filename, path=path)
    else:
        raise ValueError(
            "filename {} not a recognized compressed extension: {}".format(
                filename,
                ", ".join(ZIP_FILE_EXT + TAR_FILE_EXT),
            ),
        )


def uncompress_zip_file(zip_filename: str, path: Optional[str] = None) -> str:
    head, tail = os.path.split(zip_filename)
    for ext in ZIP_FILE_EXT:
        if tail.endswith(ext):
            tail = tail[: -(len(ext))]
            break
    path = path if path else os.path.join(head, f"{tail}-zip-uncompressed")
    logger.info(f"extracting zip {zip_filename} -> {path}")
    with zipfile.ZipFile(zip_filename) as zfile:
        zfile.extractall(path=path)
    return path


def uncompress_tar_file(tar_filename: str, path: Optional[str] = None) -> str:
    head, tail = os.path.split(tar_filename)
    for ext in TAR_FILE_EXT:
        if tail.endswith(ext):
            tail = tail[: -(len(ext))]
            break

    path = path if path else os.path.join(head, f"{tail}-tar-uncompressed")
    logger.info(f"extracting tar {tar_filename} -> {path}")
    # NOTE: "r:*" mode opens both compressed (e.g ".tar.gz") and uncompressed ".tar" archives
    with tarfile.open(tar_filename, "r:*") as tfile:
        # NOTE(robinson): Mitigate against malicious content being extracted from the tar file.
        # This was added in Python 3.12
        # Ref: https://docs.python.org/3/library/tarfile.html#extraction-filters
        if sys.version_info >= (3, 12):
            tfile.extraction_filter = tarfile.tar_filter
        else:
            logger.warning(
                "Extraction filtering for tar files is available for Python 3.12 and above. "
                "Consider upgrading your Python version to improve security. "
                "See https://docs.python.org/3/library/tarfile.html#extraction-filters"
            )
        tfile.extractall(path=path)
    return path



from dataclasses import dataclass
from typing import Optional


@dataclass
class RetryStrategyConfig:
    """
    Contains all info needed for decorator to pull from `self` for backoff
    and retry triggered by exception.

    Args:
        max_retries: The maximum number of attempts to make before giving
            up. Once exhausted, the exception will be allowed to escape.
            The default value of None means there is no limit to the
            number of tries. If a callable is passed, it will be
            evaluated at runtime and its return value used.
        max_retry_time: The maximum total amount of time to try for before
            giving up. Once expired, the exception will be allowed to
            escape. If a callable is passed, it will be
            evaluated at runtime and its return value used.
    """

    max_retries: Optional[int] = None
    max_retry_time: Optional[float] = None
